@@ -16,10 +16,12 @@
 package com.google.auto.value;
 
 import static com.google.common.truth.Truth.assertThat;
+import static com.google.common.truth.TruthJUnit.assume;
 
 import com.google.auto.value.annotations.Empty;
 import com.google.auto.value.annotations.GwtArrays;
 import com.google.auto.value.annotations.StringValues;
+import com.google.common.base.StandardSystemProperty;
 import com.google.common.collect.ImmutableCollection;
 import com.google.common.collect.ImmutableList;
 import com.google.common.collect.ImmutableSet;
@@ -30,6 +32,7 @@ import java.io.ObjectStreamClass;
 import java.lang.annotation.Annotation;
 import java.lang.annotation.Retention;
 import java.lang.annotation.RetentionPolicy;
+import java.lang.reflect.Method;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
@@ -39,6 +42,7 @@ import java.util.List;
 import java.util.Set;
 import java.util.SortedSet;
 import java.util.TreeSet;
+import org.junit.AssumptionViolatedException;
 import org.junit.Test;
 import org.junit.runner.RunWith;
 import org.junit.runners.JUnit4;
@@ -66,6 +70,34 @@ public class AutoAnnotationTest {
         .testEquals();
   }
 
+  @Test
+  public void testEqualsParameterAnnotation() throws ReflectiveOperationException {
+    assume()
+        .that(Double.parseDouble(StandardSystemProperty.JAVA_SPECIFICATION_VERSION.value()))
+        .isAtLeast(8.0);
+    Class<? extends Annotation> jspecifyNullable;
+    try {
+      // We write this using .concat in order to hide it from rewriting rules.
+      jspecifyNullable =
+          Class.forName("org".concat(".jspecify.nullness.Nullable")).asSubclass(Annotation.class);
+    } catch (ClassNotFoundException e) {
+      throw new AssumptionViolatedException("No JSpecify @Nullable available", e);
+    }
+    @SuppressWarnings("GetClassOnAnnotation") // yes, I really want the implementation class
+    Class<? extends StringValues> autoAnnotationImpl = newStringValues(new String[0]).getClass();
+    Method equals = autoAnnotationImpl.getDeclaredMethod("equals", Object.class);
+    // The remaining faffing around with reflection is there because we have a Google-internal test
+    // that runs this code with -source 7 -target 7. We're really just doing this:
+    //   assertThat(equals.getAnnotatedParameterTypes()[0].isAnnotationPresent(jspecifyNullable))
+    //      .isTrue();
+    Method getAnnotatedParameterTypes = Method.class.getMethod("getAnnotatedParameterTypes");
+    Object[] annotatedParameterTypes = (Object[]) getAnnotatedParameterTypes.invoke(equals);
+    Method isAnnotationPresent =
+        annotatedParameterTypes[0].getClass().getMethod("isAnnotationPresent", Class.class);
+    assertThat(isAnnotationPresent.invoke(annotatedParameterTypes[0], jspecifyNullable))
+        .isEqualTo(true);
+  }
+
   @Test
   public void testArraysAreCloned() {
     String[] array = {"Jekyll"};
@@ -85,7 +117,7 @@ public class AutoAnnotationTest {
     assertThat(arrays.ints()).asList().containsExactly(2, 3, 5).inOrder();
     strings[0] = "Hyde";
     ints[0] = -1;
-    assertThat(ImmutableList.copyOf(arrays.strings())).containsExactly("Jekyll");
+    assertThat(arrays.strings()).asList().containsExactly("Jekyll");
     assertThat(arrays.ints()).asList().containsExactly(2, 3, 5).inOrder();
   }
 
@@ -146,6 +178,8 @@ public class AutoAnnotationTest {
 
     StringValues anAnnotation();
 
+    Class<? extends CharSequence> aClass();
+
     byte[] bytes();
 
     short[] shorts();
@@ -167,6 +201,8 @@ public class AutoAnnotationTest {
     RetentionPolicy[] enums();
 
     StringValues[] annotations();
+
+    Class<? extends CharSequence>[] classes();
   }
 
   @AutoAnnotation
@@ -182,6 +218,7 @@ public class AutoAnnotationTest {
       String aString,
       RetentionPolicy anEnum,
       StringValues anAnnotation,
+      Class<? extends CharSequence> aClass,
       byte[] bytes,
       short[] shorts,
       int[] ints,
@@ -192,7 +229,8 @@ public class AutoAnnotationTest {
       boolean[] booleans,
       String[] strings,
       RetentionPolicy[] enums,
-      StringValues[] annotations) {
+      StringValues[] annotations,
+      Class<? extends CharSequence>... classes) {
     return new AutoAnnotation_AutoAnnotationTest_newEverything(
         aByte,
         aShort,
@@ -205,6 +243,7 @@ public class AutoAnnotationTest {
         aString,
         anEnum,
         anAnnotation,
+        aClass,
         bytes,
         shorts,
         ints,
@@ -215,7 +254,8 @@ public class AutoAnnotationTest {
         booleans,
         strings,
         enums,
-        annotations);
+        annotations,
+        classes);
   }
 
   @AutoAnnotation
@@ -231,6 +271,7 @@ public class AutoAnnotationTest {
       String aString,
       RetentionPolicy anEnum,
       StringValues anAnnotation,
+      Class<? extends CharSequence> aClass,
       Collection<Byte> bytes,
       List<Short> shorts,
       ArrayList<Integer> ints,
@@ -241,7 +282,8 @@ public class AutoAnnotationTest {
       ImmutableCollection<Boolean> booleans,
       ImmutableList<String> strings,
       ImmutableSet<RetentionPolicy> enums,
-      Set<StringValues> annotations) {
+      Set<StringValues> annotations,
+      List<Class<? extends CharSequence>> classes) {
     return new AutoAnnotation_AutoAnnotationTest_newEverythingCollections(
         aByte,
         aShort,
@@ -254,6 +296,7 @@ public class AutoAnnotationTest {
         aString,
         anEnum,
         anAnnotation,
+        aClass,
         bytes,
         shorts,
         ints,
@@ -264,7 +307,8 @@ public class AutoAnnotationTest {
         booleans,
         strings,
         enums,
-        annotations);
+        annotations,
+        classes);
   }
 
   @Everything(
@@ -279,6 +323,7 @@ public class AutoAnnotationTest {
       aString = "maybe\nmaybe not\n",
       anEnum = RetentionPolicy.RUNTIME,
       anAnnotation = @StringValues("whatever"),
+      aClass = String.class,
       bytes = {5, 6},
       shorts = {},
       ints = {7},
@@ -289,7 +334,8 @@ public class AutoAnnotationTest {
       booleans = {false, true, false},
       strings = {"ver", "vers", "vert", "verre", "vair"},
       enums = {RetentionPolicy.CLASS, RetentionPolicy.RUNTIME},
-      annotations = {@StringValues({}), @StringValues({"foo", "bar"})})
+      annotations = {@StringValues({}), @StringValues({"foo", "bar"})},
+      classes = {String.class, StringBuilder.class})
   private static class AnnotatedWithEverything {}
 
   // Get an instance of @Everything via reflection on the class AnnotatedWithEverything,
@@ -310,6 +356,7 @@ public class AutoAnnotationTest {
           "maybe\nmaybe not\n",
           RetentionPolicy.RUNTIME,
           newStringValues(new String[] {"whatever"}),
+          String.class,
           new byte[] {5, 6},
           new short[] {},
           new int[] {7},
@@ -322,7 +369,9 @@ public class AutoAnnotationTest {
           new RetentionPolicy[] {RetentionPolicy.CLASS, RetentionPolicy.RUNTIME},
           new StringValues[] {
             newStringValues(new String[] {}), newStringValues(new String[] {"foo", "bar"}),
-          });
+          },
+          String.class,
+          StringBuilder.class);
   private static final Everything EVERYTHING_FROM_AUTO_COLLECTIONS =
       newEverythingCollections(
           (byte) 1,
@@ -336,6 +385,7 @@ public class AutoAnnotationTest {
           "maybe\nmaybe not\n",
           RetentionPolicy.RUNTIME,
           newStringValues(new String[] {"whatever"}),
+          String.class,
           Arrays.asList((byte) 5, (byte) 6),
           Collections.<Short>emptyList(),
           new ArrayList<Integer>(Collections.singleton(7)),
@@ -348,7 +398,10 @@ public class AutoAnnotationTest {
           ImmutableList.of("ver", "vers", "vert", "verre", "vair"),
           ImmutableSet.of(RetentionPolicy.CLASS, RetentionPolicy.RUNTIME),
           ImmutableSet.of(
-              newStringValues(new String[] {}), newStringValues(new String[] {"foo", "bar"})));
+              newStringValues(new String[] {}), newStringValues(new String[] {"foo", "bar"})),
+          ImmutableList.of(String.class.asSubclass(CharSequence.class), StringBuilder.class));
+          // .asSubclass because of pre-Java8, where otherwise we get a compilation error because
+          // the inferred type is <Class<? extends CharSequence & Serializable>>.
   private static final Everything EVERYTHING_ELSE_FROM_AUTO =
       newEverything(
           (byte) 0,
@@ -362,6 +415,7 @@ public class AutoAnnotationTest {
           "",
           RetentionPolicy.SOURCE,
           newStringValues(new String[] {""}),
+          String.class,
           new byte[0],
           new short[0],
           new int[0],
@@ -386,6 +440,7 @@ public class AutoAnnotationTest {
           "",
           RetentionPolicy.SOURCE,
           newStringValues(new String[] {""}),
+          String.class,
           ImmutableList.<Byte>of(),
           Collections.<Short>emptyList(),
           new ArrayList<Integer>(),
@@ -396,7 +451,8 @@ public class AutoAnnotationTest {
           ImmutableSet.<Boolean>of(),
           ImmutableList.<String>of(),
           ImmutableSet.<RetentionPolicy>of(),
-          Collections.<StringValues>emptySet());
+          Collections.<StringValues>emptySet(),
+          Collections.<Class<? extends CharSequence>>emptyList());
 
   @Test
   public void testEqualsAndHashCode() {
@@ -483,6 +539,7 @@ public class AutoAnnotationTest {
             + "aByte=1, aShort=2, anInt=3, aLong=-4, aFloat=NaN, aDouble=NaN, aChar='#', "
             + "aBoolean=true, aString=\"maybe\\nmaybe not\\n\", anEnum=RUNTIME, "
             + "anAnnotation=@com.google.auto.value.annotations.StringValues([\"whatever\"]), "
+            + "aClass=class java.lang.String, "
             + "bytes=[5, 6], shorts=[], ints=[7], longs=[8, 9], floats=[10.0, 11.0], "
             + "doubles=[-Infinity, -12.0, Infinity], "
             + "chars=['?', '!', '\\n'], "
@@ -492,7 +549,8 @@ public class AutoAnnotationTest {
             + "annotations=["
             + "@com.google.auto.value.annotations.StringValues([]), "
             + "@com.google.auto.value.annotations.StringValues([\"foo\", \"bar\"])"
-            + "]"
+            + "], "
+            + "classes=[class java.lang.String, class java.lang.StringBuilder]"
             + ")";
     assertThat(EVERYTHING_FROM_AUTO.toString()).isEqualTo(expected);
     assertThat(EVERYTHING_FROM_AUTO_COLLECTIONS.toString()).isEqualTo(expected);
